/*
 * mathemu_test.c
 * illustrates bug in include/linux/math-emu/op-common.h:_FP_FROM_INT()
 */
#include <linux/kernel.h>
#include <linux/module.h>

#include <asm/byteorder.h>
#if defined(__LITTLE_ENDIAN)
#define __BYTE_ORDER __LITTLE_ENDIAN
#define __BIG_ENDIAN 4321	/* math-emu does #if __BYTE_ORDER == __BIG_ENDIAN tests */
#else
#define __BYTE_ORDER __BIG_ENDIAN
#endif

#include <asm/sfp-machine.h>
#ifndef umul_ppmm
/* umul_ppmm() is not needed for int/float conversions, but must be
   defined otherwise <math-emu/soft-fp.h> breaks */
#define umul_ppmm(wh, wl, u, v)	not needed
#endif

#include <math-emu/soft-fp.h>
#include <math-emu/single.h>

MODULE_AUTHOR("Mikael Pettersson <mikpe@it.uu.se>");
MODULE_DESCRIPTION("Test math-emu int/float conversions");
MODULE_LICENSE("GPL");

static void __init mathemu_itof(int IR, union _FP_UNION_S *rd)
{
    FP_DECL_EX;
    FP_DECL_S(SR);

    FP_FROM_INT_S(SR, IR, 32, int);
    FP_PACK_SP(rd, SR);
}

static void __init mathemu_ftoi(const union _FP_UNION_S *rs, int *IR)
{
    FP_DECL_EX;
    FP_DECL_S(SB);

    FP_UNPACK_SP(SB, rs);
    FP_TO_INT_S(*IR, SB, 32, 1);
}

static void __init itof(const int *i, float *f)
{
    union _FP_UNION_S rd;
    mathemu_itof(*i, &rd);
    *f = rd.flt;
}

static void __init ftoi(const float *f, int *i)
{
    union _FP_UNION_S rs;
    rs.flt = *f;
    mathemu_ftoi(&rs, i);
}

static int __initdata itests[] = {
    /* a set of power-of-2 test vectors <= INT32_MAX, these should all
       convert without precision loss unless something is broken */
    0x40000000,
    0x20000000,
    0x10000000,
    0x08000000, /* fails due to < vs <= error in _FP_FROM_INT */
    0x04000000,
    0x02000000,
    0x01000000,
    0x00800000,
    /* test vectors around the failure point above */
    0x0f000000,
    0x0e000000,
    0x0d000000,
    0x0c000000,
    0x0b000000,
    0x0a000000,
    0x09000000,
    0x08000000,
    0x07000000,
};

static void __init test_itof(void)
{
    unsigned int i;
    union {
	float f;
	unsigned int ui;
    } res1;
    int res2;

    for (i = 0; i < sizeof(itests)/sizeof(itests[0]); ++i) {
	itof(&itests[i], &res1.f);
	ftoi(&res1.f, &res2);
	printk(KERN_INFO "%s: 0x%08x -> 0x%08x -> 0x%08x, diff %d\n",
	       __FUNCTION__, itests[i], res1.ui, res2, res2 - itests[i]);
    }
}

int __init mathemu_test_init(void)
{
    printk(KERN_INFO "mathemu_test: init\n");
    test_itof();
    return 0;
}

void __exit mathemu_test_exit(void)
{
    printk(KERN_INFO "mathemu_test: exit\n");
}

module_init(mathemu_test_init)
module_exit(mathemu_test_exit)
